#ifndef RECONCILED_SET
#define RECONCILED_SET

/*  RECONCILED_SET
 **  A data structure for holding set elements that can be
 **  quickly reconciled with other SetRecon's.
 **
 ** %M:  sets and resets NTL's random number seed often
 */

#include <unistd.h>
#include <math.h>
#include <time.h>
#include <string>
#include <NTL/matrix.h>
#include <NTL/mat_ZZ_p.h>
#include <NTL/vec_ZZ.h>
#include <NTL/ZZ.h>
#include <NTL/ZZ_p.h>
#include <NTL/vec_ZZ_p.h>
#include <NTL/vec_vec_ZZ_p.h>
#include <NTL/ZZ_pX.h>
#include <NTL/ZZ_pXFactoring.h>
#include "polynomial.h"
#include <ostream>

// RECONCILED_SET
class SetRecon
{

	public:
		// CONSTRUCTION
		SetRecon(int the_bitsize, int the_mbar);
		/* %R: 1 <= the_kk < 2^(the_bitsize-1) - 2
		   2 <= the_pp
		   %M: most data structure elements, and the ZZ_p modulus
		   %E: basic constructor
		   if client = 0, sets up as a server, otherwise sets up as a client
		 */
		~SetRecon();                        // destructor

		// copy constructor (due to the Sketch)
		SetRecon(const SetRecon& p_recon);

		SetRecon& operator=(const SetRecon& other);

		Sketch* getPoly();

		void addPoly(Sketch* poly_to_add);

		void resetPoly();

		unsigned int getEvalSize();

		// MANIPULATION
		void add_element(ZZ_p& el);                // adds an element to the set
		//	void del_element(ZZ_p el);            // deletes an element from the set - not really useful

		int Reconcile(Sketch* p_otherPoly, unsigned int& diffASize, 
			unsigned int& diffBSize);

		int Reconcile(Sketch* p_otherPoly, vec_ZZ_p& diffA, vec_ZZ_p& diffB);
		//const SetRecon &other); // returns the contents of this set - other
		// AUXILIARY
		static void setField(int bitsize, int mbar);     // sets the size of ZZ_p being used
	//	static bool test(int the_bitsize, int setCommon, int setDiff, int the_mbar, bool confirm);

	//	static void GetMemInfo(const std::string&);
		void FreePow();
		/*
		   %R:  setCommon+setDiff < 2^(the_bitsize), the_kk < 2^(the_bitsize-1) - 2
		   sleep for one second after calling before running another test (to allow sockets to close properly)
		   %E:  reconcile two random sets with setCommon common elements and setDiff differences.
		   If confirm is true, then the result of the reconciliation is checked for accuracy.
		   The reamining parameters as in the constructor.

		   note that mutual difference is not computed in the most efficient way possible, so
		   data gleanded from this procedure is suboptimal
		 */

		static long time_setup;       // the amount of time spent on the last timed operation (in seconds)
		static long time_add;       // the amount of time spent on adding operation (in seconds)
		static long num_add;
		static long time_comput;      // a secondary timer for time spent in the last operation
		static long time_degree;
	private:
		int bitsize;           // number of bits per vector
		int mbar;              // maximum number of differences to reconcile (i.e. recovery bound)
		ZZ fieldsize;          // size of the underlying filed; a prime number in this case
		int fieldbits;         // the number of bits needed to represent fieldsize
		/*  The field is divided as follows:
		 **  -(mbar-1)/2 .. mbar/2             : sample evaluations for even mbar (-mbar/2 .. (mbar-1)/2 for odd mbar)
		 **  mbar/2+1 .. 2^bit_size + mbar/2+1 : data values
		 **  2^bit_size + mbar/2+2 .. -mbar/2  : random evaluations
		 */
		ZZ_p dataStart;        // i gets mapped to dataStart+i to avoid colisions with the sample evaluations

		vec_ZZ_p samples;      // sample points on which char poly is evaluated
		vec_ZZ_p *powSamples;  // powers of the sample points in the chosen finite field
		Sketch* m_poly;           // the partition tree containing char poly evaluations

		// auxuliary procedures
		static ZZ getField(int bitsize, int mbar);
		// %E:  returns the appropriate field size (of ZZ_p) to use for the given parameter

		void construct_RHS_constant_vector(vec_ZZ_p& RF,long& rank,long& d, vec_ZZ_p&);
		/* %R:  RF is a vector of values of the rational function evaluations at the sample points
		   (i.e. RF[i] = f_i = rational function at i'th sample point
		   rank is the rank of the LHS matrix in the system to be solved
		   d is difference in set size |S_A| - |S_B|
		   %E:  Constructs the rhs for the vandrmonde matrix - i.e. the bmatrix
		 */

		void construct_vandermonde_matrix(const vec_ZZ_p& RF, const vec_ZZ_p& bmatrix, const long d, mat_ZZ_p&);
		/* %R:  RF is a vector of values of the rational function evaluations at the sample points
		   (i.e. RF[i] = f_i = rational function at i'th sample point
		   bmatrix if the right-hand-side of the linear system to be solved (size = mbar x 1)
		   d is difference in set size |S_A| - |S_B|
		   %E:  Constructs and returns a Vandermonde matrix needed to interpolate the rational
		   function; performs some row operations to optimize performance
		 */


		// FRIENDS
		friend int reconcile_aux(SetRecon* base, Sketch* polyA, Sketch* polyB,
			unsigned int& polyADegree, unsigned int& polyBDegree, vec_ZZ_p& numerator, 
			vec_ZZ_p& denominator, bool justDegree = false);

};


#endif
